#include <DS1307RTC.h>            //Library for RTC
#include <TimeLib.h>              //Prerequisite for DS1307RTC.h
#include <SPI.h>                  //Prerequisite for SD.h
#include <SD.h>                   //Library for SD Card
#include <Wire.h>                 //Necessary for LCD and ?
#include <LiquidCrystal_I2C.h>    //Library for LCDisplay
#include <Keypad.h>               //Library for Keypad
#include <stdlib.h>               //Necessary for calibration
#include <math.h>                 //Needed for log10 operation

// ===================== //
// KEYPAD INITIALIZATION //
// ===================== //

// Define the symbols on the buttons of the keypads
char hexaKeys[2][3] = 
{
  { 
    '1','2','3'      }
  ,
  { 
    '5','6','7'      }
};
byte rowPins[2] = { 0, 2 }; //connect to the row pinouts of the keypad
byte colPins[3] = { 3, 7, 8 }; //connect to the column pinouts of the keypad

// Initialize an instance of class NewKeypad
Keypad customKeypad = Keypad( makeKeymap(hexaKeys), rowPins, colPins, 2, 3);

// ==================== //
// OTHER INITIALIZATION //
// ==================== //

// Initialize I2C library
LiquidCrystal_I2C lcd(0x27,16,2);

// Globals
int hSSpin = 4;                 // CS pin for SD-card is 4
int signpin[3] = {5,6,9};       // The LED pins are 5,6 and 9, respectively
int mm_count = 0;               // Running measurement count
String runID = "NONE";          // ID of run
String fname = "NONE";          // File name (ID of run + suffix)
int cellcount = 3;              // Number of used chambers
int interval = 1;               // Interval between measurements    
int duration = 8;               // Duration of full run
long ts;                        // Starting time of run
int maxpwm[3] = {0,0,0};        // Input voltages at maximal sensor saturation
int refmsm[3] = {0,0,0};        // Output voltages at maximal sensor saturation

// ============= //
// SETUP ROUTINE //
// ============= //

void setup(){

    initPins();                 // Pin configuration
    lcd.init();                 // LCD
    lcd.backlight();            // LCD backlight
    lcd.clear();                // Set blank screen
    initSD();                   // SD Card
    configCheck();              // Config file
    timeCheck();                // Check time
    
}


// ========= //
// MAIN LOOP //
// ========= //

void loop() {

  setFileName();                                // Set file name
  controlMenu();                                // Program setup
  createCSV();                                  // Create CSV file for data records
  int mm_max = (60*duration)/interval+1;        // Calculate number of measurements
  
  // Initial round
  ts = RTC.get();                               // Record starting time
  long tf;                                      // Initialize running time
  measureOD();                                  // First measurement routine
  digitalWrite(A3,LOW);                         // Start pump
  
  // Full run
  while(mm_count < mm_max){
      
      tf = RTC.get();                           // Get time
      delay(1000);                              // Stability
      lcd.clear();                              // Clear display
      
      // Idling until next measurement
      while((tf - ts) < (long)(mm_count*interval*60-3)){
          tf = RTC.get();
          idle();
      }

      // OD measurement
      measureOD_pausePump();
  }

  // Termination
  endProgram();
  
}


// ===================== //
// MEASUREMENT FUNCTIONS //
// ===================== //

// Measurement routine
void measureOD(){

    mm_count++;                                 // Increase measurement counter
    lcd.clear();                                // Print information on measurement
    lcd.print(F("Measuring OD!"));
    lcd.setCursor(0,1);
    lcd.print(F("Data point #"));
    lcd.print(mm_count);

    // Turn on LEDs
    switch(cellcount){
      case 3:
          analogWrite(signpin[2],maxpwm[2]);    // Turn on LED of Chamber 3 
      case 2:
          analogWrite(signpin[1],maxpwm[1]);    // Turn on LED of Chamber 2
      case 1:
          analogWrite(signpin[0],maxpwm[0]);    // Turn on LED of Chamber 1
    }
  
    delay(2000);                                // Important delay for full emission
    addRecord();                                
    delay(1000);                                // Stability delay

    // Turn off LEDs
    switch(cellcount){
      case 3:
          analogWrite(signpin[2],0);            // Turn off LED of Chamber 3 
      case 2:
          analogWrite(signpin[1],0);            // Turn off LED of Chamber 2
      case 1:
          analogWrite(signpin[0],0);            // Turn off LED of Chamber 1
    }
}

// Measurement - the actual reading
float measure(int apin){

    // Sensor reading - average of 10 sequential readings
    float x=0;
    for(int i=0;i<10;i++){
          x=x+((float)analogRead(apin)/10);
    }

    return(x);
}

// Measurement routine with pump off
void measureOD_pausePump(){

      digitalWrite(A3,HIGH);                    // Turn off pump
      delay(3000);                              // Wait 3 seconds for bubbles to disappear
      measureOD();                              // Measure OD
      digitalWrite(A3,LOW);                     // Start pump again
      delay(1000);                              // Stability delay
}

// Add line of measurements
void addRecord(){

    // Open result file for writing
    File result_csv;
    result_csv = SD.open(fname,FILE_WRITE);

    // Print measurement ID and time
    result_csv.print(mm_count);
    result_csv.print(",");
    long tn = RTC.get();
    result_csv.print(tn-ts);
    result_csv.print(",");

    // Print OD600 based on the voltage readings
    float m[3] = {0,0,0};
    m[0] = measure(0);
    result_csv.print(-log10(m[0]/refmsm[0]));
    int i;
    for(i = 1;i < cellcount;i++){
        m[i] = measure(i);
        result_csv.print(",");
        result_csv.print(-log10(m[i]/refmsm[i]));
    }

    // Print raw voltage readings
    result_csv.print(",");
    result_csv.print(m[0]);
    for(i = 1;i < cellcount;i++){
        result_csv.print(",");
        result_csv.print(m[i]);
    }
    result_csv.println();
    result_csv.close();
}


// ============== //
// ALL THE MENUS! //
// ============== //

// Menu controller. This function switches between all menus. (Settings Menu = Set.Menu)
void controlMenu(){

    char opt = mainMenu();
    while(opt!='R'){
        switch(opt){
            case 'M':
                opt = mainMenu();             // Switch to Main Menu
                break;
            case '1':
                opt = settingsMenu(0);        // Switch to Set.Menu, cursor at Date/time
                break;
            case '2':
                opt = settingsMenu(2);        // Switch to Set.Menu, cursor at Chambers used
                break;
            case '3':
                opt = settingsMenu(4);        // Switch to Set.Menu, cursor at Interval
                break;
            case '4':
                opt = settingsMenu(6);        // Switch to Set.Menu, cursor at Duration
                break;
            case '5':
                opt = settingsMenu(8);        // Switch to Set.Menu, cursor at File naming
                break;
            case '6':
                opt = settingsMenu(10);       // Switch to Set.Menu, cursor at Calibration
                break;
            case '7':
                opt = settingsMenu(11);       // Switch to Set.Menu, cursor at Test pump
                break;
            case 'D':
                opt = dateTimeMenu();         // Switch to Date/time menu
                break;
            case 'C':
                opt = chambersMenu();         // Switch to Chambers menu
                break;
            case 'I':
                opt = intervalMenu();         // Switch to Interval menu
                break;
            case 'U':
                opt = durationMenu();         // Switch to Duration menu
                break;
            case 'F':
                opt = fileNameMenu();         // Switch to File naming menu
                break;
            case 'A':
                opt = calibrationMenu();      // Switch to Calibration menu
                break;
            case 'P':
                opt = pumpMenu();             // Switch to Pump menu
                break;
        }
    }
}

// Main menu
char mainMenu(){

    lcd.clear();
    int ms = 0;                                 // Variable controlling blinking
    int menustate = 0;                          // Menustate controlling which options are displayed

    String s_c = F("> Start culture");          // Strings used for menu options display
    String s_m = F("> Edit settings");
    
    while(menustate != -1){
        while(menustate == 0){
            blinkPrint(s_c,s_m,0,&ms);
            menustate = randBlinkSwitch(0,0,1);
            if(menustate == -2){
                return('R');                    // Run program
            }
            if(menustate == -3){
                menustate = 0;                  // Stay
            }
        }
        while(menustate == 1){
            blinkPrint(s_c,s_m,1,&ms);
            menustate = randBlinkSwitch(1,0,1);
            if(menustate == -2){
                lcd.clear();
                return('1');                    // Edit settings
            }
            if(menustate == -3){
                menustate = 1;                  // Stay
            }
        }

        
    }
    
}

// Settings menu
char settingsMenu(int menustate){

    lcd.clear();
    int ms = 0;                                     // Variable controlling blinking
    String s_dat = F("> Date/time");                // Strings used for menu options display
    String s_cha = F("> Chambers used");
    String s_int = F("> Interval");
    String s_dur = F("> Duration");
    String s_fil = F("> File naming");
    String s_cal = F("> Calibration");
    String s_pum = F("> Test pump");

    // Navigating around in the Settings Menu based on 'menustate'
    while(menustate != -1){                          // Menustate (argument) is controlling which options are displayed

        while(menustate == 0){                       // NB: Inner while loops only explained for first options
            blinkPrint(s_dat,s_cha,0,&ms);           // Blinking text at upper row of display (0)
            menustate = randBlinkSwitch(0,0,1);      // Check for press-button signal (UP/DOWN/ENTER/ESC)
            if(menustate == -2){
                return('D');                         // Select current menu option (Date/time)
            }
            
        }
        
        while(menustate == 1){
          
            blinkPrint(s_dat,s_cha,1,&ms);           // Blinking text at upper row of display (1)
            menustate = randBlinkSwitch(1,0,3);
            if(menustate == -2){
                return('C');
            }
        }

        while(menustate == 2){
          
            blinkPrint(s_cha,s_int,0,&ms);
            menustate = randBlinkSwitch(2,0,3);
            if(menustate == -2){
                return('C');
            }
        }

        while(menustate == 3){
          
            blinkPrint(s_cha,s_int,1,&ms);
            menustate = randBlinkSwitch(3,2,5);
            if(menustate == -2){
                return('I');
            }
        }

        while(menustate == 4){
          
            blinkPrint(s_int,s_dur,0,&ms);
            menustate = randBlinkSwitch(4,2,5);
            if(menustate == -2){
                return('I');
            }
        }

        while(menustate == 5){
          
            blinkPrint(s_int,s_dur,1,&ms);
            menustate = randBlinkSwitch(5,4,7);
            if(menustate == -2){
                return('U');
            }
        }

        while(menustate == 6){
          
            blinkPrint(s_dur,s_fil,0,&ms);
            menustate = randBlinkSwitch(6,4,7);
            if(menustate == -2){
                return('U');
            }
        }

        while(menustate == 7){
          
            blinkPrint(s_dur,s_fil,1,&ms);
            menustate = randBlinkSwitch(7,6,9);
            if(menustate == -2){
                return('F');
            }
            
        }

        while(menustate == 8){
          
            blinkPrint(s_fil,s_cal,0,&ms);
            menustate = randBlinkSwitch(8,6,9);
            if(menustate == -2){
                return('F');
            }
            
        }

        while(menustate == 9){
          
            blinkPrint(s_fil,s_cal,1,&ms);
            menustate = randBlinkSwitch(9,8,11);
            if(menustate == -2){
                return('A');
            }
            
        }

        while(menustate == 10){
          
            blinkPrint(s_cal,s_pum,0,&ms);
            menustate = randBlinkSwitch(10,8,11);
            if(menustate == -2){
                return('A');
            }
            
        }

        while(menustate == 11){
          
            blinkPrint(s_cal,s_pum,1,&ms);
            menustate = randBlinkSwitch(11,10,11);
            if(menustate == -2){
                return('P');
            }
            
        }

        while(menustate == -3){                       // If escape was pressed
          
            return('M');                              // Return to Main Menu
            
        }
    }
}

// Date and Time Menu
char dateTimeMenu(){

    lcd.clear();
    while(true){
        printDT();                                    // Print date and time
        if(menuKey() == 4){
            return('1');
        }
    }
}

// File name
char chambersMenu(){

    int pre = cellcount;                              // Initial number of chambers
    lcd.clear();
    while(true){
        lcd.print("Chambers used: ");                 // Display information
        lcd.print(cellcount);
        lcd.setCursor(0,0);
        switch(menuKey()){                            // Check for button input
            case 1:
                if(cellcount<3){
                    cellcount++;                      // Increase number of chambers (max 3) (UP)
                }
                break;
            case 2:
                if(cellcount>1){
                    cellcount--;                      // Decrease number of chambers (min 1) (DOWN)
                }
                break;
            case 3:
                return('2');                          // Return initial number (ESC)
            case 4:
                cellcount = pre;
                return('2');                          // Return selected number (ENTER)
        }
    }
}

// Interval Menu
char intervalMenu(){

    int pre = interval;                               // Initial interval
    lcd.clear();
    while(true){
        lcd.print("Minutes: ");                       // Display information
        zeroPrint(interval);
        lcd.setCursor(0,0);
        switch(menuKey()){                            // Check for button input
            case 1:
                if(interval<99){
                    interval++;                       // Increase interval (max 99 minutes) (UP)
                }
                break;
            case 2:
                if(interval>1){
                    interval--;                       // Decrease interval (min 1 minute) (DOWN)
                }
                break;
            case 3:
                return('3');                          // Return initial interval (ESC)
            case 4:
                interval = pre;
                return('3');                          // Return new interval (ENTER)
        }
    }
}

// Duration Menu
char durationMenu(){

    int pre = duration;                               // Initial duration
    lcd.clear();
    while(true){
        lcd.print("Hours: ");                         // Display information
        zeroPrint(duration);
        lcd.setCursor(0,0);
        switch(menuKey()){                            // Check for button input
            case 1:
                if(duration<99){
                    duration++;                       // Increase duration (max 99 minutes) (UP)
                }
                break;
            case 2:
                if(duration>1){
                    duration--;                       // Decrease duration (min 1 minute) (DOWN)
                }
                break;
            case 3:
                return('4');                          // Return initial interval (ESC)
            case 4:
                duration = pre;
                return('4');                          // Return new interval (ENTER)
        }
    }
}

// File name
char fileNameMenu(){

    lcd.clear();
    while(true){
        lcd.setCursor(0,0);                           // Display informaiton
        lcd.print("File name:");
        lcd.setCursor(0,1);
        lcd.print(fname);
        if(menuKey() == 4){
            return('5');                              // Return to Settings Menu (ESC)
        }
    }
}

// Run Calibration
char calibrationMenu(){

    //Temporary Calibration shift (option to not proceed)
    lcd.clear();                                     // Display informaiton
    lcd.setCursor(0,0);
    lcd.print(F("Press + to start"));
    lcd.setCursor(0,1);
    lcd.print(F("calibration"));
    
    while(menuKey()!=3){
        if(menuKey()==4){
            return('6');                             // Return to Settings Menu (ESC), no calibration
        }
    }
    
    runCalibration();                                // Run calibration routine, if ENTER was pressed
    writeToConfig();                                 // Save calibration data to configuration file
}

//Test menu
char pumpMenu(){

    lcd.clear();                                      // Display information
    lcd.setCursor(0,0);
    lcd.print(F("Press escape to"));
    lcd.setCursor(0,1);
    lcd.print(F("stop the pump"));

    digitalWrite(A3,LOW);                             // Start pump
    
    while(true){
        if(menuKey() == 4){
            digitalWrite(A3,HIGH);                    // Turn pump off (ESC)
            return('7');                              // Return to Settings Menu
        }
    }
}


// =========== //
// CALIBRATION //
// =========== //

// Run calibration
void runCalibration(){

    // Calibration idling display
    lcd.clear();
    lcd.print(F("Calibrating..."));
    lcd.setCursor(0,1);
    lcd.print(F("Wait approx. 20s"));

    // Calibrate LEDs
    maxpwm[0] = calibrate(0);
    maxpwm[1] = calibrate(1);
    maxpwm[2] = calibrate(2);

    // Record sensor readings after calibration
    refmsm[0] = getVolt(maxpwm[0],0);
    int i;
    for(i = 1;i < cellcount;i++){
        refmsm[i] = getVolt(maxpwm[i],i);
    }

    // Report success
    lcd.clear();
    lcd.print(F("Calibration"));
    lcd.setCursor(0,1);
    lcd.print(F("completed"));
    delay(1500);
}

//Calibration routine
int calibrate(int apin){

  // Initialization
  int iter = 0;                       // Iteration count
  int x_cur = 255;                    // Current input PWM
  int x_next = 0;                     // Next input PWM
  int sz = 256;                       // Step length
  float tolmin = 3;                   // Min tolerance for convergence. Should be tuned with care
  float tolmax = 60;                  // Max tolerance for convergence. Should be tuned with care
  bool converged = false;             // Convergence status

  // Read maximal value
  float voltmax = getVolt(255,apin);  // Record maximal sensor reading
  float v_cur = voltmax;              // Current sensor reading
  float v_next = 0;                   // Next sensor reading

  // Calibration core - not deeply explained
  while(converged!=true){
    
    iter++;                           // Update iteration count
    sz = max(sz/2,1);                 // Reduce step length
        
    // Attempt step-up
    x_next = min(x_cur + sz,255);
    v_next = getVolt(x_next,apin);
    if(abs(v_next - v_cur) > 2){
      x_cur = x_next;
      v_cur = v_next;
    }

    // Attempt step-down
    else{
      x_cur = max(x_cur - sz,0);
      v_cur = getVolt(x_cur,apin);
    }

    // Check for convergence
    if((abs(voltmax - v_cur) > tolmin) && ((abs(voltmax - v_cur) < tolmax))){
      converged = true;
      return(x_cur);
    }
    
    // Avoid infinite loop
    if(iter > 30){
      return (-1);
    }
    
  }
  
}

// Get output voltage based on fixed input PWM
float getVolt(int PWM,int apin){

    analogWrite(signpin[apin],PWM);         // Input fixed PWM for fixed voltage to test
    delay(500);                             // Very important stability delay for full LED emission
    float volt = measure(apin);             // Sensor reading
    return (volt);
}


// ====================== //
// NAVIGATION AND DISPLAY //
// ====================== //

// Read menukey
int menuKey(){
    char customKey = customKeypad.getKey();//get the key value
    switch(customKey){
        case '1':
            return(1);                    //Arrow up
            break;
        case '5':
            return(2);                    //Arrow down
            break;
        case '2':
            return(3);                    //Enter
            break;
        case '3':
            return(4);                    //Escape
            break;
        default:
            return(-1);
    }
}

// Navigation in blinking screen
int randBlinkSwitch(int a,int b,int c){

    int readKey = menuKey();
    switch(readKey){
        case 1:
            return(b);          // Move up in display
            break;
        case 2:
            return(c);          // Move down
            break;
        case 3:
            return(-2);         // Select current option
            break;
        case 4:
            return(-3);         // One step backwards
            break;
        default:
            return(a);          // No action
            break;
    }
    
}

// Blinking feature
void blinkPrint(String t1, String t2, int pos, int *ms){

    // Print both menu options in display
    if(*ms < 10){
      lcd.setCursor(0,0);
      lcd.print(t1);
      lcd.setCursor(0,1);
      lcd.print(t2);
      *ms+=1;
    }

    // Print only one menu options in display (blink effect)
    else if(*ms < 25){
      if(*ms == 10){
          lcd.clear();
      }
      lcd.setCursor(0,1-pos);
      if(pos == 0){
          lcd.print(t2);
      }
      else{
          lcd.print(t1);
      }
      *ms+=1;
    }

    // Restart blinking process
    else{
        *ms = 0;
        lcd.clear();
    }
    
}

// Print date and time on display
void printDT(){

    tmElements_t tm;                // Time instance
  
    if(RTC.read(tm)){
        lcd.print("Date: ");
        zeroPrint(tm.Day);          // Print day
        lcd.print("/");
        zeroPrint(tm.Month);        // Print month
        lcd.print("/");
        zeroPrint(tmYearToCalendar(tm.Year)-2000);  // Print year
        lcd.setCursor(0,1);
        lcd.print("Time: ");
        zeroPrint(tm.Hour);         // Print hour
        lcd.print(":");
        zeroPrint(tm.Minute);       // Print minute
        lcd.print(":");
        zeroPrint(tm.Second);       // Print second
        lcd.setCursor(0,0);
    }
    else{
        lcd.print(F("NO TIME"));    // In case RTC not working, print error
    }
}

// Print numbers properly (always with two digits)
void zeroPrint(int a){
    if(a < 10){
        lcd.print("0");
    }
    lcd.print(a);
}


// ================ //
// STATUS FUNCTIONS //
// ================ //

// Idle between measurements
void idle(){

     lcd.setCursor(0,0);
     lcd.print(F("Measurements"));
     lcd.setCursor(0,1);
     lcd.print(F("completed: "));
     lcd.print(mm_count);
     delay(1000);
}

// Termination after final measurement
void endProgram(){

    digitalWrite(A3,HIGH);                    // Stop pump
    delay(1000);                              // Stability delay
    while(menuKey()!=3){                      // Display completion status until ENTER is pressed
       lcd.clear();
        lcd.print(F("Done measuring"));
        lcd.setCursor(0,1);
        lcd.print(F("Press+ to return"));
    }
}


// =============== //
// SETUP FUNCTIONS //
// =============== //

// Initialize pins
void initPins(){

    pinMode(hSSpin,INPUT_PULLUP);         // CS-pin (4) for SD-card
    pinMode(signpin[0],OUTPUT);           // LED-pin 0 (5)
    pinMode(signpin[1],OUTPUT);           // LED-pin 1 (6)
    pinMode(signpin[2],OUTPUT);           // LED-pin 2 (9)
    pinMode(A0,INPUT);                    // Signal-pin 0 (A0)
    pinMode(A1,INPUT);                    // Signal-pin 1 (A1)
    pinMode(A2,INPUT);                    // Signal-pin 2 (A2)
    pinMode(A3,OUTPUT);                   // Relay control (A3)
    digitalWrite(A3,HIGH);                // Immediately deactivate pump
}

// SD card initialization
void initSD(){
  
   if (!SD.begin(hSSpin)){                // Display error message if SD didn't initialize
      lcd.clear();
      lcd.print(F("Error: SD card"));
      lcd.setCursor(0,1);
      lcd.print(F("not initialized!"));
      lcd.setCursor(0,0);
      delay(3000);
      return;
   }
}

// Configuration file check
void configCheck(){

   File od_config;
  
   if(!SD.exists(F("odconfig.txt"))){    // Display error message if config file wasn't found
      lcd.clear();
      lcd.print(F("No configuration"));
      lcd.setCursor(0,1);
      lcd.print(F("found!"));
      lcd.setCursor(0,0);
      delay(3000);
      return;
   }
   else{
      od_config = SD.open(F("odconfig.txt"),FILE_READ);   // Open config file for reading
      if(od_config){
          runID = "";
          int line = 0;
          while(od_config.available()){
            char a = od_config.read();
            if(a!='+'){
              switch(line){
                case 1:
                  runID = runID + a;                      // Read run ID
                  break;
                case 3:
                  cellcount = a-48;                       // Read number of chambers
                  break;
                case 5:
                  maxpwm[0] = maxpwm[0]*10 + a-48;        // Read max PWM of first LED
                  break;
                case 7:
                  maxpwm[1] = maxpwm[1]*10 + a-48;        // Read max PWM of first LED
                  break;
                case 9:
                  maxpwm[2] = maxpwm[2]*10 + a-48;        // Read max PWM of third LED
                  break;
                case 11:
                  refmsm[0] = refmsm[0]*10 + a-48;        // Read max reading of first sensor
                  break;
                case 13:
                  refmsm[1] = refmsm[1]*10 + a-48;        // Read max reading of second sensor
                  break;
                case 15:
                  refmsm[2] = refmsm[2]*10 + a-48;        // Read max reading of third sensor
                  break;
              }
            }
            else{
              line++;
            }
          }
          od_config.close();
       }
       else{                                    // Display error message if config file was found but not opened
          lcd.print(F("Failed to open"));
          lcd.setCursor(0,1);
          lcd.print(F("odconfig.txt"));
          lcd.setCursor(0,0);
          delay(3000);
          return;
       }
   }
}

// Overwrite configuration file
void writeToConfig(){

   File od_config;
   
   if(SD.exists(F("odconfig.txt"))){
      SD.remove(F("odconfig.txt"));
   }

   // Update all config data
   od_config = SD.open(F("odconfig.txt"),FILE_WRITE);
   if(od_config){
       String k;
       od_config.println("+" + runID + "+");
       od_config.print("+");
       od_config.print(cellcount);
       od_config.println("+");
       od_config.print("+");
       od_config.print(maxpwm[0]);
       od_config.println("+");
       od_config.print("+");
       od_config.print(maxpwm[1]);
       od_config.println("+");
       od_config.print("+");
       od_config.print(maxpwm[2]);
       od_config.println("+");
       od_config.print("+");
       od_config.print(refmsm[0]);
       od_config.println("+");
       od_config.print("+");
       od_config.print(refmsm[1]);
       od_config.println("+");
       od_config.print("+");
       od_config.print(refmsm[2]);
       od_config.println("+");
       od_config.close();
   }
}

// Check for time
void timeCheck(){

    tmElements_t tm;            // Time instance
    
    if(!RTC.read(tm)){          // Set default time (25th of August 2016, 09:00:00), if no saved time is available
        tm.Hour = 9;
        tm.Minute = 0;
        tm.Second = 0;
        tm.Day = 25;
        tm.Month = 8;
        tm.Year = CalendarYrToTm(2016);
        if(!RTC.write(tm)){
            if(!RTC.chipPresent()){             // Display error message if RTC failed to load
                lcd.clear();
                lcd.print(F("Failed to load"));
                lcd.setCursor(0,1);
                lcd.print(F("DS1307RTC"));
                delay(3000);
            }
        }
    }
    
}

// Set file name
void setFileName(){

    short found = 0;
    int i,j;
    for(i = 0;i < 10;i++){
        for(j = 0;j < 10;j++){
            fname = runID + i + j + ".csv";   // File name based on run ID + unique integer
            if(!SD.exists(fname)){            // Searching for available unique integer
                found = 1;
                break;
            }
        }
        if(found == 1){                       // Unique integer found and chosen for file name
            break;
        }
    }
}

// Create CSV file for data storage
void createCSV(){   

    // Open record file
    File result_csv;
    result_csv = SD.open(fname,FILE_WRITE);

    // Print headline:
    result_csv.print("No.,Time_elapsed,");      // Measurement ID and time

    // OD600
    int i;
    for(i = 0;i < (cellcount-1);i++){
       result_csv.print(String(runID + "_0" + i+1 + "_OD,"));
    }
    result_csv.print(String(runID + "_0" + cellcount + "_OD,"));

    // Raw voltage
    for(i = 0;i < (cellcount-1);i++){
       result_csv.print(String(runID + "_0" + i+1 + "_raw,"));
    }
    result_csv.println(String(runID + "_0" + cellcount + "_OD,"));

    // Print reference data as second line:
    result_csv.print("Calibr.,-");
    for(i = 0;i < cellcount;i++){
       result_csv.print(",-");
    }

    // Reference voltages
    for(i = 0;i < cellcount;i++){
       result_csv.print(",");
       result_csv.print(refmsm[i]);
    }
    result_csv.println();
    result_csv.close();
  
}
